module SALTBase

using LinearAlgebra
using CatViews
using ForwardDiff: derivative  # used in gain.jl
using AbbreviatedTypes
using Printf: @printf

export LinearSolverData

# Each concrete SALT solver package (e.g., MaxwellSALT) should define a concrete subtype of
# LinearSolverData to store the information (e.g., the Maxwell operator A) needed by the
# specific linear Maxwell solver to use in the SALT solver package.
abstract type LinearSolverData end

# Each concrete SALT solver package (e.g., MaxwellSALT) should extend the following Base
# functions:
#
# - similar(lsd:LinearSolverData)
# - size(lsd::LinearSolverData)

# Each concrete SALT solver package (e.g., MaxwellSALT) should implement a concrete subtype of
# LinearSolverData and extend the following functions to be used in lasing.jl and nonlasing.jl:
#
# init_lsd!(lsd::LinearSolverData, ω::Number, ε::AbsVecNumber)
# linsolve!(x::AbsVecComplexF, lsd::LinearSolverData, b::AbsVecNumber)
# linsolve_transpose!(x::AbsVecComplexF, lsd::LinearSolverData, b::AbsVecNumber)
# linapply!(b::AbsVecComplexF, lsd::LinearSolverData, x::AbsVecNumber)
# abs2_interp!(abs2ψ::AbsVecReal, lsd::LinearSolverData, ψ::AbsVecNumber)
function init_lsd! end  # initialize lsd for current ω and ε
function linsolve! end  # x = A⁻¹ b
function linsolve_transpose! end  # x = A⁻ᵀ b
function linapply! end  # b = A x
function abs2_interp! end  # abs2ψ = abs2.(interpolate(ψ))

include("util.jl")
include("gain.jl")
include("lasing.jl")
include("nonlasing.jl")
include("anderson.jl")
include("switching.jl")
include("simulation.jl")

end # module
