# Define variables used throughout the package and functions to initialize them.

export GainProfile
export gen_gain_2lv, hole_burning!

## Two-level atoms
# Generate the default gain curve of SALT, which describes two-level atoms.
# For multiple classes of atoms describing inhomogeneous broadening, for which ωₐ and γ˔ are
# vectors, use gen_gain_2lv.(ωₐ, γ˔).
gen_gain_2lv(ωₐ::Real, γ˔::Real) = ω::Number -> γ˔ / (ω - ωₐ + im * γ˔)  # scalar
gen_gain′_2lv(ωₐ::Real, γ˔::Real) = ω::Real -> -γ˔ / (ω - ωₐ + im * γ˔)^2  # scalar

# Below, we define the formula for |γ|².  This can be easily calculated from γ, but the
# users may accidentally evaluate γ(ω) for complex ω and take its squared absolute value.
# That will give a wrong result, because the imaginary part of ω will be added to iγ in the
# denominator when calculating the absolute value.  To prevent this unfortunate accident
# from happening, we define a formula for |γ|² that accepts only real ω and ask the users to
# use it.
gen_abs2gain_2lv(ωₐ::Real, γ˔::Real) = ω::Real -> γ˔^2 / ((ω - ωₐ)^2 + γ˔^2)  # scalar; note ω is real for this
gen_abs2gain′_2lv(ωₐ::Real, γ˔::Real) = ω::Real -> -2γ˔^2 * (ω - ωₐ) / ((ω - ωₐ)^2 + γ˔^2)^2  # scalar; note ω is real for this

# Below, allow vectors and matrices to be PETSc ones if their sizes are 3×(# of grid points).
# If their sizes are the number of modes, keep them Julia vectors.

# About avoiding overtyping container-type variables:
#
# Below, for read-only container-type variables, I specify neither the container type nor
# eltype concretely.  [What are the examples of read-only container-type variables?]
#
# However, for variables into which I write, again I don't specify the container type
# concretely (so that I can allow various container type, such as Julia vectors and PETSc
# vectors), but I do specify the eltype, because otherwise I get InexactError that complains
# about incompatible element type (e.g., when attempting to store ComplexF64 in Float64).
#
# The same principle applies to writing functions as well as types.  If the function changes
# the contents of some variable, that variable is where I write, so its eltype should be
# specified to avoid InexactError.  For example, in create_A!, the argument A is where the
# output is stored, so it had better have ComplexF64 as the eltype, so I type it as
# AbsMatComplexF.  On the other hand, the argument CC is read-only, so it is typed AbsMatNumber.

# Parameters defining the SALT problem
# Consider including CC to gp, if I am really going to use ωₐ for PML for all modes.
mutable struct GainProfile{VF<:AbsVecFloat,G,G′,AG,AG′,W}  # VF can be PETSc vectors
    gain::Vector{G}  # gain curve
    gain′::Vector{G′}  # derivative of gain curve
    abs2gain::Vector{AG}  # squared absolute value of gain curve
    abs2gain′::Vector{AG′}  # derivative of squared absolute value of gain curve
    D₀::VF  # pump strength; length(D₀) = number of DOFs
    wt::Vector{W}  # function that calculates contribution of D₀ to kth atomic class; D₀ₖ = wt.(D₀)

    function GainProfile{VF,G,G′,AG,AG′,W}(gain::AbsVec{G},
                                           gain′::AbsVec{G′},
                                           abs2gain::AbsVec{AG},
                                           abs2gain′::AbsVec{AG′},
                                           D₀::AbsVecReal,
                                           wt::AbsVec{W}) where {VF<:AbsVecFloat,G,G′,AG,AG′,W}
        K = length(wt)  # number of atomic classes
        length(gain)==length(gain′)==length(abs2gain)==length(abs2gain′)==K ||
            @error "length.((gain,gain′,abs2gain,abs2gain′)) = $(length.((gain,gain′,abs2gain,abs2gain′))) should be all same."

        return new(gain, gain′, abs2gain, abs2gain′, D₀, wt)
    end
end

function GainProfile(gain::AbsVec, gain′::AbsVec, abs2gain::AbsVec, abs2gain′::AbsVec, D₀::AbsVecReal,
                     wt::AbsVec=(K=length(gain); [(d::Real->d/K) for k=1:K]))  # default wt: even distribution
    D₀_new = similar(D₀,Float)
    copyto!(D₀_new, D₀)

    return GainProfile{typeof(D₀_new),eltype.((gain,gain′,abs2gain,abs2gain′,wt))...}(gain, gain′, abs2gain, abs2gain′, D₀_new, wt)
end

# Convenience constructors for single atomic class
GainProfile(gain::Any, gain′::Any, abs2gain::Any, abs2gain′::Any, D₀::AbsVecReal) =
    GainProfile([gain], [gain′], [abs2gain], [abs2gain′], D₀)
GainProfile(gain::Any, gain′::Any, abs2gain::Any, abs2gain′::Any, N::Integer) =
    GainProfile(gain, gain′, abs2gain, abs2gain′, VecFloat(undef,N))

# Constructors using automatic differentiation
function GainProfile(gain::AbsVec,
                     D₀::AbsVecReal,
                     wt::AbsVec=(K=length(gain); [(d::Real->d/K) for k=1:K]))  # default wt: even distribution
    gain′ = [ω->(g(x) = derivative(x->real(γ(x+ω)),x) + im * derivative(x->imag(γ(x+ω)),x); g(0)) for γ = gain]
    abs2gain = [ω->abs2(γ(ω)) for γ = gain]
    abs2gain′ = [ω->derivative(abs2γ, ω) for abs2γ = abs2gain]

    return GainProfile(gain, gain′, abs2gain, abs2gain′, D₀, wt)
end

# Convenience constructors for single atomic class
GainProfile(gain::Any, D₀::AbsVecReal) = GainProfile([gain], D₀)
GainProfile(gain::Any, N::Integer) = GainProfile(gain, VecFloat(undef,N))


Base.length(gp::GainProfile) = length(gp.wt)  # number of atomic classes

# Evaluate 1 + hole-burning term = 1 + ∑|γaψ|².
function hole_burning!(hb::AbsVecNumber,  # output
                       abs2gain::Any,  # function that takes ω and produces |γ(ω)|²
                       ω::AbsVecReal,  # vector of lasing frequencies
                       a²::AbsVecReal,  # vector of squared amplitudes of unnormalized eigenmodes
                       abs2ψ::AbsVec{<:AbsVecReal})  # vector of absolute squares of normalized eigenmodes
    hb .= 1  # initialize
    for m = 1:length(a²)
        if a²[m] ≠ 0
            # @info "a²[$m] = $(a²[m]), ‖ψ[$m]‖ = $(norm(ψ[m]))"
            hb .+=  (abs2gain(ω[m]) * a²[m]) .* abs2ψ[m]
        end
    end
    # @info "‖hb‖ = $(norm(hb))"

    return nothing
end
